"""A module for representing Pauli operators.

This module provides a `Pauli` enum that represents the four Pauli operators:
I, X, Y, and Z. These operators are fundamental to quantum computing and are
used to describe the state of a qubit.

The `Pauli` enum provides a convenient way to work with Pauli operators in
Python. It includes methods for multiplication, string representation, and
color representation for plotting.

Typical usage example:

  >>> from surfq.pauli import Pauli
  >>>
  >>> p1 = Pauli.X
  >>> p2 = Pauli.Y
  >>> p3 = p1 * p2
  >>> print(p3)
  Z
"""

from enum import Enum
from functools import cached_property
from typing import override


class Pauli(Enum):
    """A class representing a single-qubit Pauli operator.

    This enum represents the four Pauli operators: I, X, Y, and Z. These
    operators are fundamental to quantum computing and are used to describe the
    state of a qubit.

    Attributes:
        I: The identity operator.
        X: The Pauli-X operator.
        Y: The Pauli-Y operator.
        Z: The Pauli-Z operator.
    """

    I = 0
    Z = 1
    X = 2
    Y = 3

    def __mul__(self, other: "Pauli") -> "Pauli":
        """Define Pauli Mulitplication."""
        return Pauli(self.value ^ other.value)

    @override
    def __repr__(self) -> str:
        match self:
            case Pauli.I:
                return "I"
            case Pauli.Z:
                return "Z"
            case Pauli.X:
                return "X"
            case Pauli.Y:
                return "Y"

    @override
    def __str__(self) -> str:
        match self:
            case Pauli.I:
                return "I"
            case Pauli.Z:
                return "Z"
            case Pauli.X:
                return "X"
            case Pauli.Y:
                return "Y"

    @cached_property
    def color(self) -> str:
        """Define color per Pauli operator for plotting."""
        match self:
            case Pauli.I:
                return "#FFFFFF"  # white for identity
            case Pauli.Z:
                return "#D15567"  # muted purple, complements red/pink
            case Pauli.X:
                return "#E17C88"  # muted red, in tone with syndrome reds
            case Pauli.Y:
                return "#F28EBF"  # softer pink, between X and Z



"""A module for representing a view of a surface code lattice.

This module provides a `LatticeView` class that represents a view of a
surface code lattice. A lattice view is a subset of the qubits in a lattice,
and it can be used to apply quantum gates to specific qubits.

Typical usage example:

  >>> from surfq import Lattice
  >>>
  >>> lattice = Lattice(3)
  >>> view = lattice[0, 0]
  >>> view.X()
"""

from typing import final

import numpy as np
from numpy.typing import NDArray

from .pauli import Pauli
from .plotting import plot_lattice

QubitMask = NDArray[np.uint64]

QubitAxisIndex = int | slice | list[int] | QubitMask

QubitIndex = (
    QubitAxisIndex | tuple[QubitAxisIndex, QubitAxisIndex] | list[tuple[int, int]]
)


@final
class LatticeView:
    """A class representing a view of a surface code lattice.

    A lattice view is a subset of the qubits in a lattice, and it can be used
    to apply quantum gates to specific qubits.

    Attributes:
        L: The size of the original LxL lattice.
        n: The number of qubits in the original lattice.
        paulis: The Pauli operators applied on the original lattice's qubits.
        tableau: The tableau of the stabilizers of the original lattice.
        mask: A mask of indexes that selects the qubits in the view prior to operations.
    """

    def __init__(
        self,
        L: int,
        paulis: NDArray[np.uint8],
        tableau: NDArray[np.uint8],
        qubits: QubitIndex,
    ):
        """Initialise a new lattice view.

        Args:
            L: The size of the original LxL lattice.
            paulis: The Pauli operators applied on the original lattice's qubits.
            tableau: The tableau of the stabilizers.
            qubits: The qubits in the view.
        """
        self.L = L
        self.n = L**2
        self.paulis = paulis
        self.tableau = tableau
        self.mask = self._build_mask(qubits)

    def _build_mask(self, qubits: QubitIndex) -> QubitMask:
        """
        Transform qubit indices into a valid mask.

        Qubit indices can be specified using either single- or double-axis notation:

        - Single-axis: Qubits are indexed over the full lattice, i.e., [0, n)
        - Double-axis: Qubits are indexed along x and y axes, i.e., [0, L) for each axis

        Each index may be one of:
        - int: A single qubit within the respective range
        - list[int]: Multiple qubits within the respective range
        - slice: Multiple qubits within the respective range
        - np.ndarray: Multiple qubits within the respective range
        - list[tuple[int, int]]: Explicit list of (x, y) coordinates

        The lattice is read from left to right and top to bottom.

        Examples:
        - Single qubit:
            - lattice[6]           : 7th qubit in the lattice
            - lattice[2, 5]        : Qubit at column 3, row 6

        - Slicing:
            - lattice[:]            : All qubits in the lattice
            - lattice[:, :]         : All qubits in the lattice
            - lattice[6, :]         : All qubits in column 7
            - lattice[:, 2]         : All qubits in row 3
            - lattice[::2, 1::2]    : Qubits at odd rows and even columns

        - Lists and arrays:
            - lattice[[1, 5, 6]]               : Qubits 2, 6, and 7
            - lattice[np.array([1, 5, 6])]     : Qubits 2, 6, and 7
            - lattice[[1, 5, 6], 2]            : Qubits 2, 6, and 7 in row 3
            - lattice[:, np.array([1, 5, 6])]  : Qubits 2, 6, and 7 in each column

        - Coordinates list:
            - lattice[[(1, 0), (5, 2)]] : Qubits at coordinates (1, 0) and (5, 2)
        """
        if isinstance(qubits, tuple):
            xs = qubits[0]
            if isinstance(xs, slice):
                start = 0 if xs.start is None else int(xs.start)  # pyright: ignore[reportAny]
                stop = self.L if xs.stop is None else int(xs.stop)  # pyright: ignore[reportAny]
                step = 1 if xs.step is None else int(xs.step)  # pyright: ignore[reportAny]
                xs = np.arange(start, stop, step, dtype=np.uint64)

            ys = qubits[1]
            if isinstance(ys, slice):
                start = 0 if ys.start is None else int(ys.start)  # pyright: ignore[reportAny]
                stop = self.L if ys.stop is None else int(ys.stop)  # pyright: ignore[reportAny]
                step = 1 if ys.step is None else int(ys.step)  # pyright: ignore[reportAny]
                ys = np.arange(start, stop, step, dtype=np.uint64)

            xs_len = 1 if isinstance(xs, int) else len(xs)
            ys_len = 1 if isinstance(ys, int) else len(ys)

            return (np.tile(xs, ys_len) + np.repeat(ys, xs_len) * self.L).astype(
                np.uint64
            )
        elif isinstance(qubits, slice):
            start = 0 if qubits.start is None else qubits.start  # pyright: ignore[reportAny]
            stop = self.n if qubits.stop is None else qubits.stop  # pyright: ignore[reportAny]
            step = 1 if qubits.step is None else qubits.step  # pyright: ignore[reportAny]
            return np.arange(start, stop, step, dtype=np.uint64)
        elif isinstance(qubits, int):
            return np.array([qubits], dtype=np.uint64)
        elif isinstance(qubits, list):
            if all(isinstance(item, int) for item in qubits):
                return np.array(qubits, dtype=np.uint64)
            else:
                return np.array([x + y * self.L for x, y in qubits])  # pyright: ignore[reportUnknownVariableType]
        else:
            return qubits

    def H(self):
        """Apply a Hadamard gate to the qubits in the view."""
        self.paulis[self.mask] = (self.paulis[self.mask] >> 1) + (
            self.paulis[self.mask] << 1
        ) % 4

        mask = self.mask
        self.tableau[:, mask], self.tableau[:, mask + self.n] = (
            self.tableau[:, mask + self.n].copy(),
            self.tableau[:, mask].copy(),
        )
        self.tableau[:, -1] ^= np.bitwise_xor.reduce(
            self.tableau[:, mask] & self.tableau[:, mask + self.n], axis=1
        )
        return self

    def S(self):
        """Apply an S gate to the qubits in the view."""
        mask = self.mask
        self.tableau[:, mask + self.n] ^= self.tableau[:, mask]

        self.tableau[:, -1] ^= np.bitwise_xor.reduce(
            self.tableau[:, mask] & self.tableau[:, mask + self.n], axis=1
        )
        # self.tableau[:, -1] ^= self.tableau[:, mask] & self.tableau[:, mask + self.n]
        return self

    def CX(self, target: "LatticeView"):
        """Apply a CNOT gate to the qubits in the view.

        Args:
            target: The target qubits for the CNOT gate.

        Raises:
            ValueError: If the number of control and target qubits do not match.
        """
        c_mask = self.mask
        t_mask = target.mask
        if len(c_mask) != len(t_mask):
            raise ValueError(
                f"dimension mismatch between control and target indexes: \
                {self.mask} and {target.mask}"
            )
        self.tableau[:, t_mask] ^= self.tableau[:, c_mask]
        self.tableau[:, self.n + c_mask] ^= self.tableau[:, self.n + t_mask]
        self.tableau[:, -1] ^= np.bitwise_xor.reduce(
            self.tableau[:, c_mask]
            & self.tableau[:, self.n + t_mask]
            & (self.tableau[:, t_mask] ^ self.tableau[:, self.n + c_mask] ^ 1),
            axis=1,
        )
        return self

    def Z(self):
        """Apply a Z gate to the qubits in the view."""
        self.paulis[self.mask] ^= Pauli.Z.value
        return self.S().S()

    def X(self):
        """Apply an X gate to the qubits in the view."""
        return self.H().Z().H()

    def Y(self):
        """Apply a Y gate to the qubits in the view."""
        return self.X().Z()

    def show(self):
        """Plot the original lattice."""
        return plot_lattice(self.L, self.tableau, self.paulis)



"""A module for representing a surface code lattice.

This module provides a `Lattice` class that represents a surface code lattice.
It is used to create a lattice of a given size, and it provides methods for
accessing and modifying the qubits in the lattice.

Typical usage example:

  >>> from surfq import Lattice
  >>>
  >>> lattice = Lattice(3)
  >>> lattice.qubits[0, 0].X()
  >>> lattice.show()
"""

from typing import final

import numpy as np
from numpy.typing import NDArray

from .lattice_view import LatticeView, QubitIndex
from .pauli import Pauli
from .plotting import plot_lattice


@final
class Lattice:
    """A class representing a surface code lattice.

    This class is used to create a surface code lattice of a given size.

    Attributes:
        L: The size of the lattice.
        n: The number of qubits in the lattice.
        tableau: The tableau of the stabilizers.
        stabilisers_coords: The coordinates of the stabilizers.
        paulis: The Pauli operators currently applied on the qubits.
    """

    def __init__(self, L: int):
        """Initialise a new surface code lattice.

        Args:
            L: The size of the LxL lattice.

        Raises:
            ValueError: If the lattice size is not a positive odd integer.
        """
        if L <= 0:
            raise ValueError("n must be a positive integer")
        elif L % 2 != 1:
            raise ValueError("n should be odd")
        self.L = L
        self.n = L**2
        self.tableau, self.stabilisers_coords = self.create_stabilisers(L)
        self.paulis: NDArray[np.uint8] = np.array(
            [Pauli.I.value for _ in range(self.n)], dtype=np.uint8
        )

    @property
    def X_stabilisers(self) -> tuple[NDArray[np.uint8], NDArray[np.float64]]:
        """Returns the X stabilizers of the lattice."""
        return self.tableau[: (self.n - 1) // 2], self.stabilisers_coords[
            : (self.n - 1) // 2
        ]

    @property
    def Z_stabilisers(self):
        """Returns the Z stabilizers of the lattice."""
        return self.tableau[(self.n - 1) // 2 :], self.stabilisers_coords[
            (self.n - 1) // 2 :
        ]

    @staticmethod
    def create_stabilisers(L: int):
        """Create the stabilizers for a surface code lattice.

        Args:
            L: The size of the lattice.

        Returns:
            A tuple containing the tableau of the stabilizers and the
            coordinates of the stabilizers.
        """
        n = L**2
        stabilisers: list[NDArray[np.uint8]] = []
        coordinates: list[tuple[float, float]] = []

        # Set up inner X stabilisers
        for row in range(L - 1):
            for col in range((L - 1) // 2):
                i = row * L
                j = 2 * col + (0 if row % 2 == 0 else 1)
                mask = i + j + np.array([0, 1, L, L + 1])
                stabilisers.append(
                    np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
                )
                coordinates.append((j + 0.5, row + 0.5))

        # Set up top X stabilisers
        for col in range((L - 1) // 2):
            j = 2 * col + 1
            mask = j + np.array([0, 1])
            stabilisers.append(
                np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
            )
            coordinates.append((j + 0.5, -0.5))

        # Set up bottom X stabilisers
        for col in range((L - 1) // 2):
            j = 2 * col
            mask = L * (L - 1) + j + np.array([0, 1])
            stabilisers.append(
                np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
            )
            coordinates.append((j + 0.5, L - 1 + 0.5))

        # Set up inner Z stabilisers
        for row in range(L - 1):
            for col in range((L - 1) // 2):
                i = row * L
                j = 2 * col + (0 if row % 2 == 1 else 1)
                mask = n + i + j + np.array([0, 1, L, L + 1])
                stabilisers.append(
                    np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
                )
                coordinates.append((j + 0.5, row + 0.5))

        # Set up left Z stabilisers
        for row in range((L - 1) // 2):
            i = 2 * row * L
            mask = n + i + np.array([0, L])
            stabilisers.append(
                np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
            )
            coordinates.append((-0.5, 2 * row + 0.5))

        # Set up right Z stabilisers
        for row in range((L - 1) // 2):
            i = (2 * row + 1) * L
            mask = n + i + L - 1 + np.array([0, L])
            stabilisers.append(
                np.isin(np.arange(n * 2 + 1), mask).astype(np.uint8),
            )
            coordinates.append((L - 1 + 0.5, 2 * row + 1 + 0.5))

        return np.array(stabilisers, dtype=np.uint8), np.array(
            coordinates, dtype=np.float64
        )

    def _validate_qubit(self, qubits: QubitIndex):
        if isinstance(qubits, list):
            qubits = np.array(qubits, dtype=np.uint64)
        if isinstance(qubits, int):
            if not 0 <= qubits < self.n:
                raise ValueError(
                    f"invalid out-of-range qubit: \
                        index must be 0 <= qubit < {self.n}"
                )
        elif isinstance(qubits, slice):
            start = 0 if qubits.start is None else qubits.start  # pyright: ignore[reportAny]
            stop = self.n if qubits.stop is None else qubits.stop  # pyright: ignore[reportAny]
            if not (0 <= start < self.n) or not (0 <= stop <= self.n):
                raise ValueError(
                    f"invalid out-of-range qubit slice {qubits}: \
                        start and end must be 0 <= qubit < {self.n}"
                )
        elif isinstance(qubits, np.ndarray):
            if not np.issubdtype(qubits.dtype, np.integer):
                raise TypeError(
                    f"qubit array must contain integers, got {qubits.dtype}"
                )
            if not np.all((0 <= qubits) & (qubits < self.n)):
                raise ValueError(
                    f"invalid out-of-range qubits in array {qubits}: \
                        all elements must satisfy 0 <= qubit < {self.n}"
                )
        else:
            if isinstance(qubits[0], int) and not 0 <= qubits[0] < self.L:
                raise ValueError(
                    f"invalid out-of-range qubit x-coordinate: \
                        must be 0 <= qubit < {self.L}"
                )
            if isinstance(qubits[0], slice):
                start = 0 if qubits[0].start is None else int(qubits[0].start)  # pyright: ignore[reportAny]
                stop = self.L if qubits[0].stop is None else int(qubits[0].stop)  # pyright: ignore[reportAny]
                if not (0 <= start < self.L) or not (0 <= stop <= self.L):
                    raise ValueError(
                        f"invalid out-of-range qubit x-coordinate slice {qubits}: \
                            must be 0 <= qubit < {self.L}"
                    )

            # y-coordinate
            if isinstance(qubits[1], int) and not 0 <= qubits[1] < self.L:
                raise ValueError(f" {self.L}")
            if isinstance(qubits[1], slice):
                start = 0 if qubits[1].start is None else int(qubits[1].start)  # pyright: ignore[reportAny]
                stop = self.L if qubits[1].stop is None else int(qubits[1].stop)  # pyright: ignore[reportAny]
                if not (0 <= start < self.L) or not (0 <= stop <= self.L):
                    raise ValueError(
                        f"invalid out-of-range qubit y-coordinate slice {qubits}: \
                            must be 0 <= qubit < {self.L}"
                    )

    def __getitem__(self, qubits: QubitIndex) -> "LatticeView":
        """Return a view of the lattice.

        Args:
            qubits: The qubits to include in the view.

        Returns:
            A view of the lattice.
        """
        self._validate_qubit(qubits)
        return LatticeView(self.L, self.paulis, self.tableau, qubits)

    def show(self) -> None:
        """Plot the lattice."""
        plot_lattice(self.L, self.tableau, self.paulis)



"""A module for plotting surface code lattices.

This module provides a set of functions for plotting surface code lattices,
including the qubits, the stabilizers, and the Pauli operators.

It uses the `matplotlib` library to generate the plots, and it provides a
set of custom styles and colors to make the plots easy to read and
understand.

Typical usage example:

  >>> from surfq import Lattice
  >>>
  >>> lattice = Lattice(3)
  >>> lattice.qubits[0, 0].X()
  >>> lattice.show()
"""

import importlib.resources
from logging import warning
from pathlib import Path, PosixPath

import matplotlib as mpl
import matplotlib.font_manager as fm
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns
from matplotlib.patches import Patch, Polygon, Wedge
from numpy.typing import NDArray

from .pauli import Pauli

sns.set_style("darkgrid")
font_resource = importlib.resources.files("surfq").joinpath("fonts/Roboto-Regular.ttf")

with importlib.resources.as_file(font_resource) as font_path:
    roboto_font = fm.FontProperties(fname=PosixPath(Path(font_path)))
    fm.fontManager.addfont(str(font_path))  # pyright: ignore[reportAny]
mpl.rcParams.update(  # pyright: ignore[reportUnknownMemberType]
    {
        "font.family": roboto_font.get_name(),
        "font.size": 12,
        "grid.color": "0.5",
        "grid.linestyle": "--",
        "grid.linewidth": 0.6,
        "xtick.color": "black",
        "ytick.color": "black",
    }
)


def plot_lattice(L: int, tableau: NDArray[np.uint8], paulis: NDArray[np.uint8]) -> None:
    """Plot a surface code lattice.

    This function plots a surface code lattice, including the qubits, the
    stabilizers, and the Pauli operators.

    Args:
        L: The size of the lattice.
        tableau: The tableau of the stabilizers.
        paulis: The Pauli operators on the qubits.
    """
    n = L**2

    if L >= 20:
        warning("Lattice dimension is too big to show")
        return

    s = 60 if L <= 10 else 25
    fontsize = 10 if L <= 10 else 7

    x_grid, y_grid = np.meshgrid(np.arange(L), np.arange(L))
    grid_x, grid_y = x_grid.flatten(), y_grid.flatten()
    ax = plt.gca()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]

    def draw_polygon_or_wedge(
        indices: NDArray[np.intp],
        color: str,
    ) -> None:
        x = indices % L
        y = L - 1 - indices // L

        if len(indices) == 4:
            pts = np.column_stack((x, y))
            pts[[-2, -1]] = pts[[-1, -2]]
            patch = Polygon(
                list(pts),
                facecolor=color,
                edgecolor="k",
                linewidth=0.75,
            )
        elif len(indices) == 2:
            cx, cy = float(np.mean(x)), float(np.mean(y))
            if y[0] == y[1] == 0:
                theta1, theta2 = (180, 0)
            elif y[0] == y[1] == L - 1:
                theta1, theta2 = (0, 180)
            elif x[0] == x[1] == 0:
                theta1, theta2 = (90, -90)
            else:
                theta1, theta2 = (-90, 90)
            patch = Wedge(
                center=(cx, cy),
                r=0.5,
                theta1=theta1,
                theta2=theta2,
                facecolor=color,
                edgecolor="k",
                linewidth=0.75,
            )
        else:
            return

        ax.add_patch(patch)  # pyright: ignore[reportUnknownMemberType]

    for stab in tableau[: (n - 1) // 2]:  # pyright: ignore[reportAny]
        indices = np.where(stab[:n] == 1)[0]  # pyright: ignore[reportAny]
        if stab[-1] != 1:
            draw_polygon_or_wedge(indices, "#6188b2")
        else:
            draw_polygon_or_wedge(indices, "#D15567")

    for stab in tableau[(n - 1) // 2 :]:  # pyright: ignore[reportAny]
        indices = np.where(stab[n:-1] == 1)[0]  # pyright: ignore[reportAny]
        if stab[-1] != 1:
            draw_polygon_or_wedge(indices, "#87b1d3")
        else:
            draw_polygon_or_wedge(indices, "#E17C88")

    ax.scatter(grid_x, grid_y, s=s, color="white", edgecolor="black")  # pyright: ignore[reportUnknownMemberType]

    ax.set_aspect("equal")  # pyright: ignore[reportUnknownMemberType]
    ax.set_yticks(np.arange(L))  # pyright: ignore[reportUnknownMemberType]
    ax.set_xticks(np.arange(L))  # pyright: ignore[reportUnknownMemberType]
    ax.set_xlim(-1, L)  # pyright: ignore[reportUnknownMemberType]
    ax.set_ylim(-1, L)  # pyright: ignore[reportUnknownMemberType]
    ax.set_yticklabels(np.arange(L)[::-1])  # pyright: ignore[reportUnknownMemberType]
    ax.tick_params(axis="x", labeltop=True, labelbottom=False, labelsize=fontsize)  # pyright: ignore[reportUnknownMemberType]
    ax.tick_params(axis="y", labelsize=fontsize)  # pyright: ignore[reportUnknownMemberType]
    ax.grid(True)  # pyright: ignore[reportUnknownMemberType]

    ax.legend(  # pyright: ignore[reportUnknownMemberType]
        handles=[
            Patch(
                facecolor="#6188b2",
                edgecolor="black",
                label="X-Stabiliser",
            ),
            Patch(
                facecolor="#87b1d3",
                edgecolor="black",
                label="Z-Stabiliser",
            ),
            Patch(
                facecolor="#D15567",
                edgecolor="black",
                label="X-Stabiliser Syndrome",
            ),
            Patch(
                facecolor="#E17C88",
                edgecolor="black",
                label="Z-Stabiliser Syndrome",
            ),
        ],
        handlelength=1,
        handleheight=1,
        borderpad=0.5,
        loc="lower center",
        bbox_to_anchor=(0.5, -0.17),
        fontsize=10,
        markerscale=2.0,
        ncol=2,
    )

    for q, op in enumerate(paulis):  # pyright: ignore[reportAny]
        if not op:
            continue
        p = Pauli(op)
        x, y = q % L, L - q // L - 1
        _ = plt.scatter(x, y, color=p.color, edgecolor="black", s=s)
        _ = plt.text(  # pyright: ignore[reportUnknownMemberType]
            x + 0.1,
            y + 0.1,
            f"$\\mathrm{{{p}}}_{{{q}}}$",
            ha="left",
            va="bottom",
            fontsize=fontsize,
        )

    plt.show()  # pyright: ignore[reportUnknownMemberType]



"""Test the Lattice class."""

import numpy as np
import pytest

from surfq.lattice import Lattice
from surfq.lattice_view import LatticeView


def test_lattice_init():
    """Test Lattice initialization."""
    lattice = Lattice(3)
    assert lattice.L == 3
    assert lattice.n == 9
    assert len(lattice.tableau) == 8
    assert len(lattice.stabilisers_coords) == 8
    assert len(lattice.paulis) == 9

    with pytest.raises(ValueError):
        Lattice(0)

    with pytest.raises(ValueError):
        Lattice(2)


def test_create_stabilisers():
    """Test the create_stabilisers method."""
    tableau, coords = Lattice.create_stabilisers(3)
    assert tableau.shape == (8, 19)
    assert coords.shape == (8, 2)


def test_x_z_stabilisers():
    """Test the X and Z stabilisers properties."""
    lattice = Lattice(3)
    x_stabs, x_coords = lattice.X_stabilisers
    z_stabs, z_coords = lattice.Z_stabilisers
    assert len(x_stabs) == 4
    assert len(x_coords) == 4
    assert len(z_stabs) == 4
    assert len(z_coords) == 4


def test_lattice_getitem():
    """Test the __getitem__ method."""
    lattice = Lattice(3)
    view = lattice[0, 0]
    assert isinstance(view, LatticeView)


def test_validate_qubit():
    """Test the _validate_qubit method."""
    lattice = Lattice(5)

    # Valid cases
    lattice._validate_qubit(0)
    lattice._validate_qubit(24)
    lattice._validate_qubit(slice(0, 25))
    lattice._validate_qubit(np.array([0, 24]))
    lattice._validate_qubit((0, 0))
    lattice._validate_qubit((slice(0, 5), slice(0, 5)))

    # Invalid cases
    with pytest.raises(ValueError):
        lattice._validate_qubit(-1)
    with pytest.raises(ValueError):
        lattice._validate_qubit(25)
    with pytest.raises(ValueError):
        lattice._validate_qubit(slice(-1, 5))
    with pytest.raises(ValueError):
        lattice._validate_qubit(slice(0, 26))
    with pytest.raises(ValueError):
        lattice._validate_qubit(np.array([-1, 24]))
    with pytest.raises(ValueError):
        lattice._validate_qubit(np.array([0, 25]))
    with pytest.raises(TypeError):
        lattice._validate_qubit(np.array([0.0, 24.0]))
    with pytest.raises(ValueError):
        lattice._validate_qubit((-1, 0))
    with pytest.raises(ValueError):
        lattice._validate_qubit((0, 5))
    with pytest.raises(ValueError):
        lattice._validate_qubit((slice(-1, 5), slice(0, 5)))
    with pytest.raises(ValueError):
        lattice._validate_qubit((slice(0, 6), slice(0, 5)))



"""Test Pauli Frame."""

from surfq.pauli import Pauli


def test_single_pauli_operations():
    """Test Single-Qubit Pauli operations."""
    assert Pauli.I * Pauli.I == Pauli.I
    assert Pauli.X * Pauli.I == Pauli.X
    assert Pauli.Z * Pauli.I == Pauli.Z
    assert Pauli.Y * Pauli.I == Pauli.Y
    assert Pauli.I * Pauli.X == Pauli.X
    assert Pauli.X * Pauli.X == Pauli.I
    assert Pauli.Z * Pauli.X == Pauli.Y
    assert Pauli.Y * Pauli.X == Pauli.Z
    assert Pauli.I * Pauli.Z == Pauli.Z
    assert Pauli.X * Pauli.Z == Pauli.Y
    assert Pauli.Z * Pauli.Z == Pauli.I
    assert Pauli.Y * Pauli.Z == Pauli.X
    assert Pauli.I * Pauli.Y == Pauli.Y
    assert Pauli.X * Pauli.Y == Pauli.Z
    assert Pauli.Z * Pauli.Y == Pauli.X
    assert Pauli.Y * Pauli.Y == Pauli.I



"""Test the LatticeView class."""

import numpy as np
import pytest

from surfq.lattice import Lattice
from surfq.pauli import Pauli


@pytest.fixture
def lattice():
    """Return a 3x3 lattice."""
    return Lattice(3)


def test_build_mask(lattice: Lattice):
    """Test the _build_mask method."""
    # Single-axis indexing
    np.testing.assert_array_equal(lattice[0].mask, [0])
    np.testing.assert_array_equal(lattice[8].mask, [8])
    np.testing.assert_array_equal(lattice[:].mask, np.arange(9))
    np.testing.assert_array_equal(lattice[1:4].mask, np.arange(1, 4))
    np.testing.assert_array_equal(lattice[[0, 8]].mask, [0, 8])

    # Double-axis indexing
    np.testing.assert_array_equal(lattice[0, 0].mask, [0])
    np.testing.assert_array_equal(lattice[2, 2].mask, [8])
    np.testing.assert_array_equal(lattice[:, :].mask, np.arange(9))
    np.testing.assert_array_equal(lattice[0, :].mask, [0, 3, 6])
    np.testing.assert_array_equal(lattice[:, 0].mask, [0, 1, 2])
    np.testing.assert_array_equal(lattice[1:3, 1:3].mask, [4, 5, 7, 8])
    np.testing.assert_array_equal(lattice[[0, 2], [0, 2]].mask, [0, 2, 6, 8])

    # List of tuples
    np.testing.assert_array_equal(lattice[[(0, 0), (2, 2)]].mask, [0, 8])


def test_h_gate(lattice: Lattice):
    """Test the H gate."""
    _ = lattice[0].H()
    assert lattice.paulis[0] == Pauli.I.value


def test_s_gate(lattice: Lattice):
    """Test the S gate."""
    _ = lattice[0].S()
    assert lattice.paulis[0] == Pauli.I.value


def test_cx_gate(lattice: Lattice):
    """Test the CX gate."""
    _ = lattice[0].CX(lattice[1])
    _ = lattice[0, :].CX(lattice[1, :])
    assert lattice.paulis[0] == Pauli.I.value
    assert lattice.paulis[1] == Pauli.I.value

    with pytest.raises(ValueError):
        _ = lattice[0].CX(lattice[[1, 2]])


def test_x_gate(lattice: Lattice):
    """Test the X gate."""
    _ = lattice[0].X()
    assert lattice.paulis[0] == Pauli.X.value


def test_y_gate(lattice: Lattice):
    """Test the Y gate."""
    _ = lattice[0].Y()
    assert lattice.paulis[0] == Pauli.Y.value


def test_z_gate(lattice: Lattice):
    """Test the Z gate."""
    _ = lattice[0].Z()
    assert lattice.paulis[0] == Pauli.Z.value



"""A library for simulating the surface code.

This library provides a set of tools for simulating the behavior of the surface code.

It includes classes for representing the lattice, the qubits, and the
stabilizers, as well as functions for applying quantum gates and simulating
errors.

Typical usage example:

  >>> from surfq import Lattice
  >>>
  >>> lattice = Lattice(3)
  >>> lattice.qubits[0, 0].X()
  >>> lattice.qubits[0, 1].Z()
  >>> lattice.show()
"""

from .lattice import Lattice

__all__ = ["Lattice"]
